import inspect
import sys
import types

from seleniumpm.webelements.clickable import Clickable
from seleniumpm.webelements.element import Element


class Widget(Clickable):
    def __init__(self, driver, locator=None):
        super(Clickable, self).__init__(driver, locator)

    def dict(self):
        """
        This returns a dictionary representation of a Widget

        :return: a dict representation of a Widget
        """
        dictionary = super(Widget, self).dict()
        elements = self.get_element_attr_local()
        dictionary['elements'] = {}
        for key, element in elements.iteritems():
            dictionary['elements'][key] = element.dict()
        dictionary['methods'] = self.get_methods_local()
        return dictionary

    def wait_for_widget_load(self, timeout=None, force_check_visibility=False, check_myself=True):
        """
        This implementation is the same as Webpage.wait_for_page_load(). This is to support iFrame situations that
        is a separate page load than the outer page.

        :param timeout: (Default: 30s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :param check_myself: (Default: True) Since a Widget/Panel/IFrame is a type of Element, it also has a locator.
                             This is used for enabling/disabling adding a validation against itself. The scenario
                             where this could be used is in a situation where you want to validate the elements of this
                             container, but not the Locator of the container itself.
        :raises TimeoutException: if an element doesn't appear within timeout
        :return: self
        """
        timeout = timeout if timeout is not None else self.page_timeout
        return self.validate(timeout=timeout, force_check_visibility=force_check_visibility, check_myself=check_myself)

    def validate(self, timeout=None, force_check_visibility=True, check_myself=False):
        """
        The intention of validate is to make sure that an already loaded widget contains these elements.

        :param timeout: (Default: 10s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :param check_myself: (Default: True) Since a Widget/Panel/IFrame is a type of Element, it also has a locator.
                             This is used for enabling/disabling adding a validation against itself. The scenario
                             where this could be used is in a situation where you want to validate the elements of this
                             container, but not the Locator of the container itself.
        :raises TimeoutException: if an element doesn't appear within timeout
        :return: self
        """
        timeout = timeout if timeout is not None else self.element_timeout
        from seleniumpm.iframe import IFrame
        for element in self.get_element_attr(expand_iframe_elements=False, check_myself=check_myself):
            # Continue if the element has marked itself do_not_check=True or a Locator is not defined
            if element.do_not_check or element.locator is None:
                continue
            # Check for presence and visibility
            if force_check_visibility or element.check_visible:
                # Print to stderr a WARNING message when force_check_visibility=True and element has been marked 'invisible'
                if force_check_visibility and not element.check_visible:
                    sys.stderr.write(
                        "[WARNING] element {}={} ({}) was marked as 'invisible' but force_check_visibility=True".format(
                            element.locator.by, element.locator.value, self.__class__))
                if isinstance(element, IFrame):
                    element.validate(timeout=timeout, force_check_visibility=force_check_visibility)
                else:
                    element.wait_for_present_and_visible(timeout)
            else:
                element.wait_for_present(timeout)
        return self

    def get_element_attr(self, type=Element, override_check_visible=False, override_do_not_check=False,
                         expand_iframe_elements=False, result_type=list, check_myself=True, attr_name='widget'):
        """
        Retrieves a list of WebElements on a Widget. Optionally, you can pass in a different type (e.g. Button,
        Link, TextElement) to return only those types associated with a Widget object.

        :param type: one of the seleniumpm.webelement types (Default: seleniumpm.webelements.Element)
        :param override_check_visible: (Default: False) This overrides check for visibility. By default, widgets that are
                                        invisible, means that we assume its elements are also invisible
        :param override_do_not_check: (Default: False) This overrides check for do_not_check. By default, widgets that are
                                        marked as do_not_check, means that we assume its elements are not accessible
        :param expand_iframe_elements: (Default: False) Elements within an iFrame must be kept together in order to execute validate()
        :param result_type: (Default: list) This value can either be (list|dict). By default, we simply want a list of
                            available elements on the page. However, the dictionary version is implemented for the ability
                            to retrieve every element and sub-element on a Webpage directly from the page level
        :param check_myself: (Default: True) Since a Widget/Panel/IFrame is a type of Element, it also has a locator.
                             This is used for enabling/disabling adding a validation against itself. The scenario
                             where this could be used is in a situation where you want to validate the elements of this
                             container, but not the Locator of the container itself.
        :param attr_name: (Default: 'widget') This is for passing the attribute name from its parent when recursively
                            iterating through all of its sub-elements.
        :return: This is a list or dict of attributes of base type seleniumpm.webelements.Element
        """
        from seleniumpm.iframe import IFrame
        from seleniumpm.webelements.panel import Panel
        if result_type != list and result_type != dict:
            raise AttributeError(
                "result_type can either be 'list' (default) or 'dict', but was '{}'".format(result_type))
        elements = [] if result_type == list else {}
        temp_widgets = {}
        # Add myself if I match the expected type
        if isinstance(self, type) and (not isinstance(self, IFrame) or expand_iframe_elements) and check_myself:
            if result_type == dict:
                elements[attr_name] = self
            else:
                elements.append(self)
        for attr in dir(self):
            # This is to catch potential exceptions thrown in situations where a developer
            # decorates a method with @property and the method raises an error
            try:
                element = getattr(self, attr, None)
            except:
                element = None
            # Ensure that it is of type Element
            if isinstance(element, Element):
                # If it is a widget, then recursively drill down and get its Elements
                if isinstance(element, Widget):
                    # Check if widget is a type of iFrame and also check for visibility
                    if (element.check_visible or override_check_visible) and \
                            (not element.do_not_check or override_do_not_check) and \
                            not isinstance(element, IFrame) or \
                            ((isinstance(element, IFrame) and expand_iframe_elements)):
                        if result_type == dict:
                            temp_widgets[attr] = []
                            for key, value in element.get_element_attr(type=type,
                                                                       override_check_visible=override_check_visible,
                                                                       override_do_not_check=override_do_not_check,
                                                                       expand_iframe_elements=expand_iframe_elements,
                                                                       result_type=result_type,
                                                                       check_myself=True,
                                                                       attr_name=attr).items():
                                temp_widgets[attr].append({'key': key, 'value': value})
                        else:
                            for welement in element.get_element_attr(type=type):
                                elements.append(welement)
                    # Add the widget but not its sub-elements if invisible
                    else:
                        if result_type == dict:
                            elements[attr] = element
                        else:
                            elements.append(element)
                # Add the element if it matches the expected type (not a Widget)
                if type not in (Widget, Panel, IFrame) and isinstance(element, type) and not isinstance(element,
                                                                                                        Widget):
                    if result_type == dict:
                        elements[attr] = element
                    else:
                        elements.append(element)

        # Give non-widgets priority, hence why there is a separate loop for widgets and their elements
        for attr, values in temp_widgets.items():
            for element in values:
                if element['key'] not in elements:
                    elements[element['key']] = element['value']
                else:
                    elements["{}_{}".format(attr, element['key'])] = element['value']
        return elements

    def get_element_attr_local(self):
        """
        This is a much simpler implement of get_element_attr() in that it only returns back the locally defined
        Elements, and not any elements defined in sub-Widgets and sub-Panels.

        :return: A dict of Element types
        """
        elements = {}
        for attr in dir(self):
            element = getattr(self, attr)
            if isinstance(element, Element) and attr in self.__dict__.keys():
                elements[attr] = element
        return elements

    def get_methods_local(self):
        """
        Returns only the local methods defined for this class

        :return: a dict containing method names (keys) and a list of parameters for the method (values)
        """
        results = {}
        for attr in dir(self):
            method = getattr(self, attr)
            if type(method) == types.MethodType and \
                            method.__name__ not in ('__init__') and \
                            method.__func__ in method.im_class.__dict__.values():
                args = inspect.getargspec(method)[0][1:]
                results[method.__name__] = args
        return results

    def get_all_elements_on_widget(self):
        """
        Retrieves all the webelements that have been defined on the widget. This includes all sub-elements found on
        widgets, panels, and iframes.

        :return: A dict of all the elements
        """
        return self.get_element_attr(override_check_visible=True,
                                     override_do_not_check=True,
                                     expand_iframe_elements=True,
                                     result_type=dict)

    def __getattr__(self, name):
        """
        Overridden method so that we can also be able to directly access all webelements defined at lower level
        webelements (a.k.a. widgets, panels, and iframes)

        :param name: The name of the attribute or webelement that we expect to exist on the Webpage
        :return: The attribute value if found
        :raises AttributeError: If the attribute with given name doesn't exist
        """
        # Need to escape a potential infinite recursion for the following names
        if name in ('__members__', '__methods__'):
            return
        # Checking if the element is defined in the sub-widgets, panels, or iframes
        all_elements = self.get_all_elements_on_widget()
        if name in all_elements:
            return all_elements[name]
        else:
            raise AttributeError(
                "'{}' webpage and its widgets has no attribute '{}'. "
                "The following are valid webelements on the page:\n  - {}".format(
                    self.__class__.__name__, name, "\n  - ".join(all_elements.keys())))



from seleniumpm.webelements.widget import Widget
from seleniumpm.webelements.link import Link
from seleniumpm.locator import Locator
from selenium.webdriver.common.by import By

class WikipediaPersonal(Widget):
    """
    This is an example way of sub-organizing elements on a page. In this example, we're grouping a set of
    links located near the top-right of the Wikipedia page.
    """

    def __init__(self, driver, locator):
        super(WikipediaPersonal, self).__init__(driver, locator)
        self.talk_link = Link(driver, Locator(By.XPATH, "//li[@id='pt-anontalk']/a"))
        self.contributions_link = Link(driver, Locator(By.XPATH, "//li[@id='pt-anoncontribs']/a"))
        self.createaccount_link = Link(driver, Locator(By.XPATH, "//li[@id='pt-createaccount']/a"))
        self.login_link = Link(driver, Locator(By.XPATH, "//li[@id='pt-login']/a"))



from functools import wraps
import time

import seleniumpm.config as seleniumconfig
from seleniumpm.webpage import Webpage

def take_screenshot_on_test_error(func):
    @wraps(func)
    def newFunc(*args, **kwargs):
        try:
            func_response = func(*args, **kwargs)
        except Exception as e:
            if seleniumconfig.test_screenshot_enabled:
                funcObj = args[0]
                filename = "test_error_%s_%s" % (func.func_name, time.strftime('%Y_%m_%d-%H_%M_%S'))
                page = Webpage(funcObj.driver)
                page.take_screenshot(screenshot_name=filename)
                import sys
                exc_class, exc, tb = sys.exc_info()
                new_exc = exc_class("\n%s\nScreenshot file: %s.png" % (exc or exc_class, filename))
                raise new_exc.__class__, new_exc, tb
            raise e
        return func_response

    return newFunc



from selenium.webdriver.support.select import Select
from seleniumpm.webelements.clickable import Clickable
from seleniumpm.webelements.element import take_screenshot_on_element_error


class Dropdown(Clickable):
    def __init__(self, driver, locator):
        super(Dropdown, self).__init__(driver, locator)
        self.select = None

    @take_screenshot_on_element_error
    def instantiateSelect(self):
        if self.select == None:
            self.select = Select(self.get_webelement())

    def count_options(self):
        # TODO - Need to implement a way to get and modify the original locator to append '/option' (Assuming xpath)
        raise NotImplementedError

    def select_by_visible_text(self, txt):
        self.instantiateSelect()
        self.select.select_by_visible_text(txt)
        return self

    def select_by_index(self, index):
        self.instantiateSelect()
        self.select.select_by_index(index)
        return self



from functools import wraps
import base64
import os
import re
import time

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.action_chains import ActionChains
import seleniumpm.config as seleniumconfig
from seleniumpm.locator import Locator

# Regular expression to find numbers (both int and float) in a string
number_re = r'([\-]*\d+\.\d+|[\-]*\d+)'

def take_screenshot_on_element_error(func):
    @wraps(func)
    def newFunc(*args, **kwargs):
        try:
            # Disable Screenshot
            current_screenshot_enabled = seleniumconfig.screenshot_enabled
            if current_screenshot_enabled:
                seleniumconfig.screenshot_enabled = False
            try:
                func_response = func(*args, **kwargs)
            finally:
                # Reset screenshot
                seleniumconfig.screenshot_enabled = current_screenshot_enabled
        except Exception as e:
            if seleniumconfig.screenshot_enabled:
                funcObj = args[0]
                filename = "element_error_%s_%s" % (func.func_name, time.strftime('%Y_%m_%d-%H_%M_%S'))
                page = Element(funcObj.driver, Locator.by_xpath("//placeholder"))
                page.take_screenshot(screenshot_name=filename)
                import sys
                exc_class, exc, tb = sys.exc_info()
                new_exc = exc_class("\n%s\nScreenshot file: %s.png" % (exc or exc_class, filename))
                raise new_exc.__class__, new_exc, tb
            raise e
        return func_response

    return newFunc

class Element(object):

    def __init__(self, driver, locator):
        if not seleniumconfig.disable_check_for_selenium_webdriver and not isinstance(driver, WebDriver):
            raise AttributeError("driver was not an expected RemoteWebdriver type!")
        if not isinstance(locator, Locator) and locator is not None:
            raise AttributeError("locator was not an expected seleniumpm.Locator type!")
        self.driver = driver
        self.locator = locator

        # These attributes controls Webpage and Widget validation checks
        self.check_visible = True
        self.do_not_check = False

    def mark_invisible(self):
        self.check_visible = False
        return self

    def mark_do_not_check(self):
        self.do_not_check = True
        return self

    @take_screenshot_on_element_error
    def get_webelement(self):
        if self.locator is None:
            raise AttributeError("locator was not specified!")
        return self.driver.find_element(self.locator.by, self.locator.value)

    @take_screenshot_on_element_error
    def get_webelements(self):
        if self.locator is None:
            raise AttributeError("locator was not specified!")
        return self.driver.find_elements(self.locator.by, self.locator.value)

    def get_action_chains(self):
        return ActionChains(self.driver)

    @take_screenshot_on_element_error
    def get_text(self):
        return self.get_webelement().text

    def get_texts(self):
        web_elements = self.get_webelements()
        results = []
        for web_element in web_elements:
            results.append(web_element.text)
        return results

    @take_screenshot_on_element_error
    def get_number(self, string=None, result_index=0):
        """
        This simplifies getting a number from an element

        :param string: (Optional) Default: None - This is a string that we want to extract a number from. By default
                        this is implicitly doing an element.text operation to retrieve the string.
        :param result_index: (Optional) Default: 0 - By default, converts and returns the first matching set. This can
                                be changed to return the n'th matching set
        :return: A number that can be (i) an int, (ii) a float, or (iii) None if neither.
        """
        string = self.get_text() if string == None else string
        try:
            return self.get_int(string, result_index)
        except ValueError:
            try:
                return self.get_float(string, result_index)
            except ValueError:
                return None

    def get_numbers(self):
        """
        This simplifies getting a list of numbers from a set of web elements

        :return: A list of numbers where each item can be (i) an int, (ii) a float, or (iii) None if neither.
        """
        web_elements = self.get_webelements()
        results = []
        for web_element in web_elements:
            results.append(self.get_number(web_element.text))
        return results

    @take_screenshot_on_element_error
    def get_int(self, string=None, result_index=0):
        """
        This simplifies getting an integer from an element

        :param string: (Optional) Default: None - This is a string that we want to extract a number from. By default
                        this is implicitly doing an element.text operation to retrieve the string.
        :param result_index: (Optional) Default: 0 - By default, converts and returns the first matching set. This can
                                be changed to return the n'th matching set
        :return: an integer value
        :raises ValueError: if the element text is not an integer
        """
        string = self.get_text() if string == None else string
        results = map(int, re.findall(number_re, string))
        return results[result_index] if len(results) > 0 else None

    @take_screenshot_on_element_error
    def get_float(self, string=None, result_index=0):
        """
        This simplifies getting an float from an element

        :param string: (Optional) Default: None - This is a string that we want to extract a number from. By default
                        this is implicitly doing an element.text operation to retrieve the string.
        :param result_index: (Optional) Default: 0 - By default, converts and returns the first matching set. This can
                                be changed to return the n'th matching set
        :return: an float value
        :raises ValueError: if the element text is not an float
        """
        string = self.get_text() if string == None else string
        results = map(float, re.findall(number_re, string))
        return results[result_index] if len(results) > 0 else None

    @take_screenshot_on_element_error
    def get_attribute(self, name):
        """
        Performs a Webelement.get_attribute() and thus returns back a string

        :param name: This is the attribute that you want to retrieve (e.g. class or href) the value of
        :return: string
        """
        return self.get_webelement().get_attribute(name)

    def get_attribute_contains(self, name, value):
        """
        Returns True|False depending on whether or not the specified value is in the results. This is implemented
        specifically for matching css classes, and thus the attribute values **MUST** be space-delimited

        :param name: The attribute name to retrieve
        :param value: A specific attribute value that you want to match
        :return: True|False if the attribute value is in the list
        """
        attribute = self.get_attribute(name)
        attributes = attribute.split()
        return True if value in attributes else False

    def get_attribute_is(self, name, value):
        """
        Returns True|False depending on whether or not the specified value matches the returned attribute value

        :param name: The attribute name to retrieve
        :param value: A specific attribute value that you want to match
        :return: True|False if the attribute value is an exact match
        """
        return True if self.get_attribute(name) == value else False

    @take_screenshot_on_element_error
    def is_displayed(self):
        return self.get_webelement().is_displayed()

    @take_screenshot_on_element_error
    def is_enabled(self):
        return self.get_webelement().is_enabled()

    @take_screenshot_on_element_error
    def is_selected(self):
        return self.get_webelement().is_selected()

    @take_screenshot_on_element_error
    def is_present(self, timeout=None):
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            self.wait_for_present(timeout)
            return True
        except:
            return False

    @take_screenshot_on_element_error
    def is_visible(self, timeout=None):
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            self.wait_for_visible(timeout)
            return True
        except:
            return False

    @take_screenshot_on_element_error
    def is_present_and_visible(self, timeout=None, present_timeout=None, visible_timeout=None):
        present_timeout = present_timeout if present_timeout is not None else self.element_timeout
        visible_timeout = visible_timeout if visible_timeout is not None else self.element_timeout
        return self.is_present(timeout if timeout is not None else present_timeout) and \
               self.is_visible(timeout if timeout is not None else visible_timeout)

    @take_screenshot_on_element_error
    def set_focus(self):
        return self.scroll_into_view()

    @take_screenshot_on_element_error
    def scroll_into_view(self):
        self.driver.execute_script("arguments[0].scrollIntoView();", self.get_webelement())
        return self

    @take_screenshot_on_element_error
    def move_to_element(self):
        return self.get_action_chains().move_to_element(self.get_webelement()).build().perform()

    @take_screenshot_on_element_error
    def hover_over(self):
        self.move_to_element()

    @take_screenshot_on_element_error
    def get_html(self):
        """
        Retrieves the inner-html of an element

        :return: str representing the inner-html of an element
        """
        return self.get_webelement().get_attribute("innerHTML").encode("utf-8")

    @take_screenshot_on_element_error
    def wait_for_selected(self, timeout=None):
        if self.locator is None:
            raise AttributeError("locator was not specified!")
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            WebDriverWait(self.driver, timeout).until(
                EC.element_to_be_selected((self.locator.by, self.locator.value)))
        except TimeoutException as e:
            e.message = "TimeoutException waiting for selected {}={} with timeout={}s ({})".format(self.locator.by, self.locator.value, timeout, self.__class__)
            e.msg = e.message
            raise e
        return self

    @take_screenshot_on_element_error
    def wait_for_present(self, timeout=None):
        if self.locator is None:
            raise AttributeError("locator was not specified!")
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            WebDriverWait(self.driver, timeout).until(
                EC.presence_of_element_located((self.locator.by, self.locator.value)))
        except TimeoutException as e:
            e.message = "TimeoutException waiting for present {}={} with timeout={}s ({})".format(self.locator.by, self.locator.value, timeout, self.__class__)
            e.msg = e.message
            raise e
        return self

    @take_screenshot_on_element_error
    def wait_for_visible(self, timeout=None):
        if self.locator is None:
            raise AttributeError("locator was not specified!")
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            WebDriverWait(self.driver, timeout).until(
                EC.visibility_of_element_located((self.locator.by, self.locator.value)))
        except TimeoutException as e:
            e.message = "TimeoutException waiting for visible {}={} with timeout={}s ({})".format(self.locator.by, self.locator.value, timeout, self.__class__)
            e.msg = e.message
            raise e
        return self

    @take_screenshot_on_element_error
    def wait_for_present_and_visible(self, timeout=None, present_timeout=None, visible_timeout=None):
        present_timeout = present_timeout if present_timeout is not None else self.element_timeout
        visible_timeout = visible_timeout if visible_timeout is not None else self.element_timeout
        self.wait_for_present(timeout if timeout is not None else present_timeout)
        self.wait_for_visible(timeout if timeout is not None else visible_timeout)

    def take_screenshot(self, screenshot_dir=None, screenshot_name=None, debug_logger_object=None):
        """
        Allows you to take a screenshot of the current page.

        :param screenshot_dir: (Default: './screenshots') The directory path for the screenshots
        :param screenshot_name: (Default: "screenshot_%s" % time.strftime('%Y_%m_%d-%H_%M_%S')) The file name excluding
                                the type
        :param debug_logger_object: (Default: None) Ability to reference your own debugger object. I am assuming there
                                    is a debug(msg) method, in which this method will write to.
        :return: screenshot_name
        """
        screenshot_name = "screenshot_%s" % time.strftime(
            '%Y_%m_%d-%H_%M_%S') if screenshot_name is None else screenshot_name
        screenshot_dir = seleniumconfig.screenshot_dir if screenshot_dir is None else screenshot_dir
        debug_logger_object = seleniumconfig.debug_logger_object if debug_logger_object is None else debug_logger_object
        filename = "%s/%s.png" % (screenshot_dir, screenshot_name)

        # Ensure that path exists, otherwise create it
        if not os.path.exists(screenshot_dir):
            os.makedirs(screenshot_dir)
        # Debugging information
        if debug_logger_object is not None:
            debug_logger_object.debug("Saving ScreenShot at %s" % filename)
        else:
            print "[DEBUG] Saving Screenshot at %s" % filename

        base64_data = self.driver.get_screenshot_as_base64()
        screenshot_data = base64.decodestring(base64_data)
        screenshot_file = open(filename, "w")
        screenshot_file.write(screenshot_data)
        screenshot_file.close()

        return screenshot_name
        return self

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def dict(self):
        """
        This returns a dictionary representation of a Element

        :return: a dict representation of a Element
        """
        return {
            'type': self.__class__.__name__,
            'package': self.__module__,
            'locator': self.locator.dict(),
            'check_visible': self.check_visible,
            'do_not_check': self.do_not_check
        }

    @property
    def page_timeout(self):
        return self.get_page_timeout()

    @property
    def element_timeout(self):
        return self.get_element_timeout()

    def get_page_timeout(self):
        return seleniumconfig.page_timeout_in_sec

    def get_element_timeout(self):
        return seleniumconfig.element_timeout_in_sec

    def start_timer(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will "start" the timer and set driver.start_time = time.time().

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The start time
        """
        attr = "start_time" if type is None else "{}_start_time".format(type)
        setattr(self.driver, attr, time.time())
        return getattr(self.driver, attr)

    def stop_timer(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will "stop" the timer and set driver.end_time = time.time().

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The end time
        """
        start_attr = "start_time" if type is None else "{}_start_time".format(type)
        end_attr = "end_time" if type is None else "{}_end_time".format(type)
        duration_attr = "duration_time" if type is None else "{}_duration_time".format(type)

        # Return None if haven't started a timer
        if not hasattr(self.driver, start_attr) or getattr(self.driver, start_attr) == 0:
            return None

        setattr(self.driver, end_attr, time.time())
        start_time = getattr(self.driver, start_attr)
        end_time = getattr(self.driver, end_attr)
        setattr(self.driver, duration_attr, end_time - start_time)
        return end_time

    def get_split_time(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will return a duration between driver.start_time and now. If a timer was not started, then return 0.

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The duration between driver.start_time and time.time(). Otherwise, 0
        """
        attr = "start_time" if type is None else "{}_start_time".format(type)
        if not hasattr(self.driver, attr) or getattr(self.driver, attr) == 0:
            return 0
        return time.time() - getattr(self.driver, attr)

    def get_duration(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will return a duration between driver.start_time and now. This method will also call stop_timer(). If a timer
        was already stopped, then do not call stop_timer() again; instead return the previous duration. If a timer was
        not started, then return 0.

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The duration between driver.start_time and driver.end_time. Otherwise, 0
        """
        start_attr = "start_time" if type is None else "{}_start_time".format(type)
        end_attr = "end_time" if type is None else "{}_end_time".format(type)
        duration_attr = "duration_time" if type is None else "{}_duration_time".format(type)
        if not hasattr(self.driver, start_attr):
            return 0
        if not hasattr(self.driver, end_attr):
            self.stop_timer(type=type)
        if getattr(self.driver, end_attr) < getattr(self.driver, start_attr):
            self.stop_timer(type=type)
        return getattr(self.driver, duration_attr)

    def reset_timer(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will set driver.start_time = 0, driver.end_time = 0, and driver.duration_time = 0

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: self
        """
        start_attr = "start_time" if type is None else "{}_start_time".format(type)
        end_attr = "end_time" if type is None else "{}_end_time".format(type)
        duration_attr = "duration_time" if type is None else "{}_duration_time".format(type)
        setattr(self.driver, start_attr, 0)
        setattr(self.driver, end_attr, 0)
        setattr(self.driver, duration_attr, 0)
        return self



from selenium.webdriver.common.by import By
from seleniumpm.webpage import Webpage
from seleniumpm.webelements.textfield import TextField
from seleniumpm.locator import Locator


class GooglePage(Webpage):
    """
    This is an Google page that extends SeleniumPM WebPage. This class acts as a container for the different
    WebElements on the page that an engineer may want to interact with.
    """

    def __init__(self, driver, url=None):
        super(GooglePage, self).__init__(driver, url)
        self.search_field = TextField(driver, Locator.by_name('q'))

    def get_result_links(self):
        """
        Returns a list of links from a Google search.
        :return: Returns a list of links from a Google search.
        """
        links = []
        elements = self.driver.find_elements(By.XPATH, "//h3[contains(@class, 'r')]/a")
        for element in elements:
            links.append(element.get_attribute("href"))
        return links



# highlevel selenium debug flag
debug = False

# Page and element timeouts for Selenium
page_timeout_in_sec = 30
element_timeout_in_sec = 10

def reset_timeouts():
    global page_timeout_in_sec
    global element_timeout_in_sec
    page_timeout_in_sec = 30
    element_timeout_in_sec = 10

# Screenshot variables
screenshot_dir = "./screenshots"
screenshot_enabled = True
test_screenshot_enabled = True

# Selenium Checks
disable_check_for_selenium_webdriver = False

# Debugging configs
debug_logger_object = None



from seleniumpm.webelements.textelement import TextElement
from seleniumpm.webelements.element import take_screenshot_on_element_error
import time


class TextField(TextElement):
    def __init__(self, driver, locator):
        super(TextField, self).__init__(driver, locator)

    @take_screenshot_on_element_error
    def send_keys(self, txt):
        self.get_webelement().send_keys(txt)
        return self

    @take_screenshot_on_element_error
    def type(self, txt):
        return self.send_keys(txt)

    @take_screenshot_on_element_error
    def send_keys_delayed(self, txt, delay=0.2):
        element = self.get_webelement()
        for i in txt:
            time.sleep(delay)
            element.send_keys(i)
        return self

    @take_screenshot_on_element_error
    def type_delayed(self, txt, delay=0.2):
        return self.send_keys_delayed(txt=txt, delay=delay)

    @take_screenshot_on_element_error
    def clear(self):
        self.get_webelement().clear()
        return self

    @take_screenshot_on_element_error
    def submit(self):
        self.get_webelement().submit()



from seleniumpm.webelements.widget import Widget


class Panel(Widget):
    """
    The Panel type is a construct for describing a specific section of a page. This is conceptually the same
    as a Widget type, but a Panel appears to be a generally accepted term. Again, A Panel/Widget is way to
    organize the elements/operations of a section of a Webpage. A Panel can contain an Element type, or even
    another (sub-)Panel/Widget.

    Since a Panel/Widget represents a specific section of a page (e.g. a Header or Footer), then it also
    has a Location associated with it.
    """

    def __init__(self, driver, locator=None):
        super(Panel, self).__init__(driver=driver, locator=locator)



from seleniumpm.webelements.clickable import Clickable
from seleniumpm.webelements.element import take_screenshot_on_element_error


class Checkbox(Clickable):
    def __init__(self, driver, locator):
        super(Checkbox, self).__init__(driver, locator)

    @take_screenshot_on_element_error
    def select(self):
        if not self.is_selected():
            self.get_webelement().click()

    @take_screenshot_on_element_error
    def unselect(self):
        if self.is_selected():
            self.get_webelement().click()



from seleniumpm.locator import Locator
from seleniumpm.webelements.element import Element
from seleniumpm.webelements.panel import Panel
from seleniumpm.webelements.element import take_screenshot_on_element_error


class IFrame(Panel):
    def __init__(self, driver, locator=None):
        super(IFrame, self).__init__(driver=driver, locator=locator)
        self.iframe_load_duration_time = 0

    def get_html(self, switch_in=True):
        """
        Retrieves the html of the entire page

        :param switch_in: (Default: True) This is used for automatically switching into and out of an iFrame context
        :return: a str of the entire page html
        """
        try:
            if switch_in:
                self.switch_in()
            return Element(self.driver, Locator.by_xpath("//html")).get_html()
        finally:
            if switch_in:
                self.switch_out()

    def wait_for_iframe_load(self, timeout=None, force_check_visibility=True, check_myself=True):
        """
        An IFrame wait_for_iframe_load() takes into account that you have to switch_in() to an iFrame to actually
        validate the inner-page.

        :param timeout: (Default: 30s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :param check_myself: (Default: True) Since a Widget/Panel/IFrame is a type of Element, it also has a locator.
                             This is used for enabling/disabling adding a validation against itself. The scenario
                             where this could be used is in a situation where you want to validate the elements of this
                             container, but not the Locator of the container itself.
        :raises TimeoutException: if an element doesn't appear within timeout
        :return: self
        """
        timeout = timeout if timeout is not None else self.page_timeout
        return self.validate(timeout=timeout, force_check_visibility=force_check_visibility, check_myself=check_myself)

    def validate(self, timeout=None, force_check_visibility=False, check_myself=True):
        """
        An IFrame validate takes into account that you have to switch_in() to an iFrame to actually validate the
        inner-page.

        :param timeout: (Default: 10s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :param check_myself: (Default: True) Since a Widget/Panel/IFrame is a type of Element, it also has a locator.
                             This is used for enabling/disabling adding a validation against itself. The scenario
                             where this could be used is in a situation where you want to validate the elements of this
                             container, but not the Locator of the container itself.
        :raises TimeoutException: if an element doesn't appear within timeout
        :return: self
        """
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            self.start_timer(type="iframe_load")
            self.switch_in() if check_myself else None
            return super(IFrame, self).validate(timeout=timeout,
                                                force_check_visibility=force_check_visibility,
                                                check_myself=check_myself)
        finally:
            self.switch_out() if check_myself else None
            self.stop_timer(type="iframe_load")

    @take_screenshot_on_element_error
    def switch_in(self):
        """
        This is to support switching to an iFrame for smart validations of all Element's on a page

        TODO - This implementation lacks support of multiple embedded iFrame's (e.g. an iFrame within another iFrame)

        :return: self
        """
        if self.locator:
            self.driver.switch_to.frame(self.driver.find_element(by=self.locator.by, value=self.locator.value))
        return self

    @take_screenshot_on_element_error
    def switch_out(self):
        """
        This is to support switching out of an iFrame for smart validations of all Element's on a page

        TODO - This implementation lacks support of multiple embedded iFrame's (e.g. an iFrame within another iFrame)

        :return: self
        """
        self.driver.switch_to.default_content()
        return self

    @property
    def iframe_load_duration(self):
        driver_iframe_load_duration_time = self.get_duration(type="iframe_load")
        if driver_iframe_load_duration_time > self.iframe_load_duration_time:
            self.iframe_load_duration_time = driver_iframe_load_duration_time
        return self.iframe_load_duration_time



from seleniumpm.locator import Locator
from seleniumpm.webpage import Webpage
from seleniumpm.webelements.element import Element
from seleniumpm.webelements.textelement import TextElement
from selenium.webdriver.common.by import By

class Wikipedia(Webpage):
    """
    This is a basic Wikipedia page. For the constructor, the URL path will be used as the Webpage path
    """

    def __init__(self, driver, url):
        super(Wikipedia, self).__init__(driver, url)
        self.path = self.url.path
        self.headingText = TextElement(driver, Locator(By.ID, "firstHeading"))
        self.bodyText = TextElement(driver, Locator(By.ID, "bodyContent"))
        self.bodyTexts = TextElement(driver, Locator(By.XPATH, "//div[@id='bodyContent']/div/p"))
        self.bogus_element = Element(driver, Locator.by_xpath("//foo")).mark_do_not_check()

    def get_text(self):
        """Retrieves all the words in the body
        """
        txt = ""
        paragraphs = self.bodyTexts.get_webelements()
        for paragraph in paragraphs:
            txt += " {}".format(paragraph.text.encode('ascii', 'ignore'))
        return txt



from seleniumpm.examples.wikipedia import Wikipedia
from seleniumpm.examples.widgets.WikipediaPersonal import WikipediaPersonal
from seleniumpm.webelements.link import Link
from seleniumpm.locator import Locator


class SuperWikipedia(Wikipedia):
    """
    This is simply an example of how you could extend Webpage classes. One use-case of this could be that your
    website has a HomepageUnauthenticated and a HomepageAuthenticated; in essence, you could implement
    2 classes that represent the base Homepage class in (1) an authenticated state, and (2) an unauthenticated
    state.
    """

    def __init__(self, driver, url):
        super(SuperWikipedia, self).__init__(driver, url)
        self.mainpage_link = Link(driver, Locator.by_xpath("//li[@id='n-mainpage-description']/a"))
        self.personal_widget = WikipediaPersonal(driver, Locator.by_xpath("//div[@id='p-personal']"))



from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

import seleniumpm.config as seleniumconfig

from seleniumpm.iframe import IFrame
from seleniumpm.locator import Locator
from seleniumpm.webelements.element import Element
from seleniumpm.webelements.widget import Widget
from seleniumpm.webelements.panel import Panel

from functools import wraps
from urlparse import urlparse
import base64
import inspect
import json
import os
import re
import sys
import time
import types

url_regex = re.compile(
    r'^(?:http|ftp)s?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE)


def take_screenshot_on_webpage_error(func):
    """
    This is an annotation for automatic screenshot ability for Webpage functions. It leverages "higher-order" operations
    available through functools library; and as the function implies, allows for automatic taking of a screenshot
    when there is an error/exception thrown.
    """
    @wraps(func)
    def newFunc(*args, **kwargs):
        try:
            # Disable Screenshot
            current_screenshot_enabled = seleniumconfig.screenshot_enabled
            if current_screenshot_enabled:
                seleniumconfig.screenshot_enabled = False
            try:
                func_response = func(*args, **kwargs)
            finally:
                # Reset screenshot
                seleniumconfig.screenshot_enabled = current_screenshot_enabled
        except Exception as e:
            if seleniumconfig.screenshot_enabled:
                funcObj = args[0]
                filename = "page_error_%s_%s" % (func.func_name, time.strftime('%Y_%m_%d-%H_%M_%S'))
                page = Webpage(funcObj.driver)
                page.take_screenshot(screenshot_name=filename)
                import sys
                exc_class, exc, tb = sys.exc_info()
                new_exc = exc_class("\n%s\nScreenshot file: %s.png" % (exc or exc_class, filename))
                raise new_exc.__class__, new_exc, tb
            raise e
        return func_response

    return newFunc


class Webpage(object):
    """
    The Webpage class is intended to be the parent class for all Webpages. In principle, a Webpage is simply a
    construct that allows us to organize a set of WebElements on any given page; therefore, this class and other
    Selenium PageModel entities follows this paradigm.

    The first step to do is to write your own class that extends Webpage:

        class Google(Webpage):
            def __init__(self, driver, url=None):
                super(Google, self).__init__(driver, url)
                self.path = ""
                self.search_field = TextField(driver, Locator.by_name('q'))

    In the constructor, you do the following:

        1) Call the parent class's constructor
        2) You define 'path' attribute (default is "", so it didn't need to be specified in this example)
        3) You add a set of WebElements to your Webpage
    """

    def __init__(self, driver, url=None):
        if not seleniumconfig.disable_check_for_selenium_webdriver and not isinstance(driver, WebDriver):
            raise AttributeError("driver was not an expected RemoteWebdriver type!")
        self.driver = driver
        self.path = ""
        self.page_load_duration_time = 0
        # Check if a valid url
        if url and not url_regex.search(url.geturl() if isinstance(url, tuple) else url):
            raise AttributeError("Invalid url: '{}'".format(url))
        # Check if  url is None or has already been parsed
        if url == None or isinstance(url, tuple):
            self.url = url
        # Check if url is defined but has not been parsed
        elif url and not isinstance(url, tuple):
            self.url = urlparse(url)

    def open(self, url=None):
        """
        This method has two forms of operation:

        1) It will open whatever url that is passed into method
        2) It will use the url specified when the WebPage object was specified (Recommended)

        The latter method is the recommended approach.

        :param url: The url to open, but it is recommended that this be specified in constructor - Default: None
        :return:
        """
        if url:
            self.driver.get(url)
        elif self.url:
            url = "{}://{}:{}{}".format(
                self.url.scheme,
                self.url.hostname,
                self.url.port if self.url.port else (80 if self.url.scheme == "http" else 443),
                self.path)
            self.driver.get(url)
        else:
            raise AttributeError("Url is not defined!")
        return self

    def refresh(self):
        """Does a page refresh
        """
        self.driver.refresh()

    def reload(self):
        """Does a page refresh
        """
        self.refresh()

    def close(self):
        """Closes the browser
        """
        self.driver.close()

    def quit(self):
        """Quits the Selenium session
        """
        self.driver.quit()

    def get_title(self):
        """Returns the title of the WebPage
        """
        return self.driver.title

    def dict(self, simple=False):
        """
        This returns a dictionary representation of the Webpage

        :param simple: (Default: False) If set to True, this will return a simple representation of a Webpage containing
                        only the list of 'methods' and a list of 'accessible_elements'
        :return: a dict representation of a Webpage
        """
        dictionary = {
            'type': "{}.{}".format(self.__module__, self.__class__.__name__),
        }
        if not simple:
            dictionary['url'] = self.url.geturl()
            dictionary['path'] = self.path
            elements = self.get_element_attr_local()
            dictionary['elements'] = {}
            for key, element in elements.iteritems():
                dictionary['elements'][key] = element.dict()
        methods = self.get_methods_local()
        dictionary['methods'] = methods.keys() if simple else methods
        dictionary['accessible_elements'] = self.get_all_elements_on_page().keys()
        return dictionary

    def to_json(self, simple=False):
        """Returns a json string
        """
        return json.dumps(self.dict(simple=simple))

    def to_json_pp(self, simple=False):
        """Returns a pretty-print json string
        """
        return json.dumps(self.dict(simple=simple), indent=4)

    def start_timer(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will "start" the timer and set driver.start_time = time.time().

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The start time
        """
        attr = "start_time" if type is None else "{}_start_time".format(type)
        setattr(self.driver, attr, time.time())
        return getattr(self.driver, attr)

    def stop_timer(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will "stop" the timer and set driver.end_time = time.time().

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The end time
        """
        start_attr = "start_time" if type is None else "{}_start_time".format(type)
        end_attr = "end_time" if type is None else "{}_end_time".format(type)
        duration_attr = "duration_time" if type is None else "{}_duration_time".format(type)

        # Return None if haven't started a timer
        if not hasattr(self.driver, start_attr) or getattr(self.driver, start_attr) == 0:
            return None

        setattr(self.driver, end_attr, time.time())
        start_time = getattr(self.driver, start_attr)
        end_time = getattr(self.driver, end_attr)
        setattr(self.driver, duration_attr, end_time - start_time)
        return end_time

    def get_split_time(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will return a duration between driver.start_time and now. If a timer was not started, then return 0.

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The duration between driver.start_time and time.time(). Otherwise, 0
        """
        attr = "start_time" if type is None else "{}_start_time".format(type)
        if not hasattr(self.driver, attr) or getattr(self.driver, attr) == 0:
            return 0
        return time.time() - getattr(self.driver, attr)

    def get_duration(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will return a duration between driver.start_time and now. This method will also call stop_timer(). If a timer
        was already stopped, then do not call stop_timer() again; instead return the previous duration. If a timer was
        not started, then return 0.

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: The duration between driver.start_time and driver.end_time. Otherwise, 0
        """
        start_attr = "start_time" if type is None else "{}_start_time".format(type)
        end_attr = "end_time" if type is None else "{}_end_time".format(type)
        duration_attr = "duration_time" if type is None else "{}_duration_time".format(type)
        if not hasattr(self.driver, start_attr) or getattr(self.driver, start_attr) == 0:
            return 0
        if not hasattr(self.driver, end_attr):
            self.stop_timer(type=type)
        if getattr(self.driver, end_attr) < getattr(self.driver, start_attr):
            self.stop_timer(type=type)
        return getattr(self.driver, duration_attr)

    @property
    def page_load_duration(self):
        driver_page_load_duration_time = self.get_duration(type="page_load")
        if driver_page_load_duration_time > self.page_load_duration_time:
            self.page_load_duration_time = driver_page_load_duration_time
        return self.page_load_duration_time

    def reset_timer(self, type=None):
        """
        This method is part of the stop-watch set of capabilities for PageObjects and elements. This method
        will set driver.start_time = 0, driver.end_time = 0, and driver.duration_time = 0

        :param type: (Default: None) This is used for capturing different timers. If not None (e.g. 'page' or
                        'element') then all timer attributes will be prefixed with '<type>_'
                        (e.g. driver.page_start_time, page_end_time, and page_duration_time)
        :return: self
        """
        start_attr = "start_time" if type is None else "{}_start_time".format(type)
        end_attr = "end_time" if type is None else "{}_end_time".format(type)
        duration_attr = "duration_time" if type is None else "{}_duration_time".format(type)
        setattr(self.driver, start_attr, 0)
        setattr(self.driver, end_attr, 0)
        setattr(self.driver, duration_attr, 0)
        return self

    def get_current_url(self):
        """Returns the current page url
        """
        return self.driver.current_url

    @property
    def current_url(self):
        """Returns the current page url
        """
        return self.get_current_url()

    @property
    def page_timeout(self):
        return self.get_page_timeout()

    @property
    def element_timeout(self):
        return self.get_element_timeout()

    def get_page_timeout(self):
        return seleniumconfig.page_timeout_in_sec

    def get_element_timeout(self):
        return seleniumconfig.element_timeout_in_sec

    @take_screenshot_on_webpage_error
    def get_html(self):
        """
        Retrieves the html of the entire webpage

        :return: a str of the entire page html
        """
        return Element(self.driver, Locator.by_xpath("//html")).get_html()

    @take_screenshot_on_webpage_error
    def wait_for_title(self, title, timeout=None):
        """This could be used similar to a wait_for_page_load() if the page title can uniquely identify
        different pages or states of the page. Google Search works like this.

        :param title: The title to search for (case sensitive)
        :param timeout: The number of seconds to wait - Default: 10s
        :raises TimeoutException: if the title does not appear within timeout period
        """
        timeout = timeout if timeout is not None else self.element_timeout
        WebDriverWait(driver=self.driver, timeout=timeout).until(EC.title_contains(title))
        return self

    @take_screenshot_on_webpage_error
    def wait_for_page_load(self, timeout=None, force_check_visibility=False):
        """
        This method "waits for page load" by checking that all expected objects are both present and visible on the
        page. This is similar to validate() operation except that sometimes certain pages take a long time to load.
        Typically the threshold is 30sec, but this is configurable.

        :param timeout: (Default: 30s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :return: self if everything is successful
        :raises TimeoutException: if an element doesn't appear within timeout
        """
        timeout = timeout if timeout is not None else self.page_timeout
        self.start_timer(type="page_load")
        self.validate(timeout=timeout, force_check_visibility=force_check_visibility)
        self.stop_timer(type="page_load")
        return self

    @take_screenshot_on_webpage_error
    def validate(self, timeout=None, force_check_visibility=False):
        """
        The intention of validate is to make sure that an already loaded webpage contains these elements.

        :param timeout: (Default: 10s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :raises TimeoutException: if an element doesn't appear within timeout
        """
        timeout = timeout if timeout is not None else self.element_timeout
        for element in self.get_element_attr():
            # Continue if the element has marked itself do_not_check=True
            if element.do_not_check:
                continue
            # Check for presence and visibility
            if force_check_visibility or element.check_visible:
                # Print to stderr a WARNING message when force_check_visibility=True and element has been marked 'invisible'
                if force_check_visibility and not element.check_visible:
                    sys.stderr.write(
                        "[WARNING] element {}={} ({}) was marked as 'invisible' but force_check_visibility=True".format(
                            element.locator.by, element.locator.value, self.__class__))
                if isinstance(element, IFrame):
                    element.validate(timeout=timeout, force_check_visibility=force_check_visibility)
                else:
                    element.wait_for_present_and_visible(timeout)
            else:
                element.wait_for_present(timeout)
        return self

    def is_page(self, timeout=None, force_check_visibility=False):
        """
        This is like validate() operation except that it returns a boolean True/False. The idea is to
        ask whether or not you are on a page; this is an implementation of that idea. There are of course
        other ways of checking whether you are on the right page or not (e.g. checking the page title).

        :param timeout: (Default: 30s) The number of seconds to poll waiting for an element
        :param force_check_visibility: (Default: False) Some elements can mark itself as invisible (but present) on
                                       load. The default is to respect this setting and only check for presence. Setting
                                       this to 'True' means you want to check for both present and visible.
        :return: True if validate() does not throw an exception; False otherwise
        """
        timeout = timeout if timeout is not None else self.page_timeout
        try:
            self.validate(timeout=timeout, force_check_visibility=force_check_visibility)
            return True
        except:
            return False

    def take_screenshot(self, screenshot_dir=None, screenshot_name=None, debug_logger_object=None):
        """
        Allows you to take a screenshot of the current page.

        :param screenshot_dir: (Default: './screenshots') The directory path for the screenshots
        :param screenshot_name: (Default: "screenshot_%s" % time.strftime('%Y_%m_%d-%H_%M_%S')) The file name excluding
                                the type
        :param debug_logger_object: (Default: None) Ability to reference your own debugger object. I am assuming there
                                    is a debug(msg) method, in which this method will write to.
        :return: screenshot_name
        """
        screenshot_name = "screenshot_%s" % time.strftime(
            '%Y_%m_%d-%H_%M_%S') if screenshot_name is None else screenshot_name
        screenshot_dir = seleniumconfig.screenshot_dir if screenshot_dir is None else screenshot_dir
        debug_logger_object = seleniumconfig.debug_logger_object if debug_logger_object is None else debug_logger_object
        filename = "%s/%s.png" % (screenshot_dir, screenshot_name)

        # Ensure that path exists, otherwise create it
        if not os.path.exists(screenshot_dir):
            os.makedirs(screenshot_dir)
        # Debugging information
        if debug_logger_object is not None:
            debug_logger_object.debug("Saving ScreenShot at %s" % filename)
        else:
            print "[DEBUG] Saving Screenshot at %s" % filename

        base64_data = self.driver.get_screenshot_as_base64()
        screenshot_data = base64.decodestring(base64_data)
        screenshot_file = open(filename, "w")
        screenshot_file.write(screenshot_data)
        screenshot_file.close()

        return screenshot_name

    def get_element_attr(self, type=Element, override_check_visible=False, override_do_not_check=False,
                         expand_iframe_elements=False, result_type=list):
        """
        Retrieves a list of WebElements on a Webpage. Optionally, you can pass in a different type (e.g. Button,
        Link, TextElement) to return only those types associated with a Webpage object.

        :param type: one of the seleniumpm.webelement types (Default: seleniumpm.webelements.Element)
        :param override_check_visible: (Default: False) This overrides check for visibility. By default, widgets that are
                                        invisible, means that we assume its elements are also invisible
        :param override_do_not_check: (Default: False) This overrides check for do_not_check. By default, widgets that are
                                        marked as do_not_check, means that we assume its elements are not accessible
        :param expand_iframe_elements: (Default: False) Elements within an iFrame must be kept together in order to execute validate()
        :param result_type: (Default: list) This value can either be (list|dict). By default, we simply want a list of
                            available elements on the page. However, the dictionary version is implemented for the ability
                            to retrieve every element and sub-element on a Webpage directly from the page level
        :return: This is a list or dict of attributes of base type seleniumpm.webelements.Element
        """
        if result_type != list and result_type != dict:
            raise AttributeError(
                "result_type can either be 'list' (default) or 'dict', but was '{}'".format(result_type))
        elements = [] if result_type == list else {}
        temp_widgets = {}
        for attr in dir(self):
            # This is to catch potential exceptions thrown in situations where a developer
            # decorates a method with @property and the method raises an error
            try:
                element = getattr(self, attr, None)
            except:
                element = None
            # Ensure that it is of type Element
            if isinstance(element, Element):
                # If it is a widget, then recursively drill down and get its Elements
                if isinstance(element, Widget):
                    # Check if widget is a type of iFrame and also check for visibility
                    if (element.check_visible or override_check_visible) and \
                            (not element.do_not_check or override_do_not_check) and \
                            not isinstance(element, IFrame) or \
                            ((isinstance(element, IFrame) and expand_iframe_elements)):
                        if result_type == dict:
                            temp_widgets[attr] = []
                            for key, value in element.get_element_attr(type=type,
                                                                       override_check_visible=override_check_visible,
                                                                       override_do_not_check=override_do_not_check,
                                                                       expand_iframe_elements=expand_iframe_elements,
                                                                       result_type=result_type,
                                                                       check_myself=True,
                                                                       attr_name=attr).items():
                                temp_widgets[attr].append({'key': key, 'value': value})
                        else:
                            for welement in element.get_element_attr(type=type):
                                elements.append(welement)
                    # Add the widget but not its sub-elements if invisible
                    else:
                        if result_type == dict:
                            elements[attr] = element
                        else:
                            elements.append(element)
                # Add the element if it matches the expected type (not a Widget)
                if type not in (Widget, Panel, IFrame) and isinstance(element, type) and not isinstance(element,
                                                                                                        Widget):
                    if result_type == dict:
                        elements[attr] = element
                    else:
                        elements.append(element)

        # Give non-widgets priority, hence why there is a separate loop for widgets and their elements
        for attr, values in temp_widgets.items():
            for element in values:
                if element['key'] not in elements:
                    elements[element['key']] = element['value']
                else:
                    elements["{}_{}".format(attr, element['key'])] = element['value']

        # Ensure that there is only 1 iframe defined on page at this time
        count_iframes = 0
        # TODO - apparently list(elements.values()) is how this should be implemented if using Python 3
        list_elements = elements.values() if result_type == dict else elements
        for element in list_elements:
            if isinstance(element, IFrame) and not element.do_not_check:
                count_iframes += 1
            if count_iframes > 1 and not expand_iframe_elements:
                raise AttributeError(
                    "There was more than 1 IFrame found on this page. This is currently not supported. "
                    "If necessary, please mark at least 1 of them using 'mark_do_not_check()'")

        return elements

    def get_element_attr_local(self):
        """
        This is a much simpler implement of get_element_attr() in that it only returns back the locally defined
        Elements, and not any elements defined in sub-Widgets and sub-Panels.

        :return: A dict of Element types
        """
        elements = {}
        for attr in dir(self):
            element = getattr(self, attr)
            if isinstance(element, Element) and attr in self.__dict__.keys():
                elements[attr] = element
        return elements

    def get_methods_local(self):
        """
        Returns only the local methods defined for this class

        :return: a dict containing method names (keys) and a list of parameters for the method (values)
        """
        results = {}
        for attr in dir(self):
            method = getattr(self, attr)
            if type(method) == types.MethodType and \
                            method.__name__ not in ('__init__') and \
                            method.__func__ in method.im_class.__dict__.values():
                args = inspect.getargspec(method)[0][1:]
                results[method.__name__] = args
        return results

    def get_all_elements_on_page(self):
        """
        Retrieves all the webelements that have been defined on the page. This includes all sub-elements found on
        widgets, panels, and iframes.

        :return: A dict of all the elements
        """
        return self.get_element_attr(override_check_visible=True,
                                     override_do_not_check=True,
                                     expand_iframe_elements=True,
                                     result_type=dict)

    def __getattr__(self, name):
        """
        Overridden method so that we can also be able to directly access all webelements defined at lower level
        webelements (a.k.a. widgets, panels, and iframes)

        :param name: The name of the attribute or webelement that we expect to exist on the Webpage
        :return: The attribute value if found
        :raises AttributeError: If the attribute with given name doesn't exist
        """
        # Need to escape a potential infinite recursion for the following names
        if name in ('__members__', '__methods__'):
            return
        # Checking if the element is defined in the sub-widgets, panels, or iframes
        all_elements = self.get_all_elements_on_page()
        if name in all_elements:
            return all_elements[name]
        else:
            raise AttributeError(
                "'{}' webpage and its widgets has no attribute '{}'. The following are valid webelements on the page:\n  - {}".format(
                    self.__class__.__name__, name, "\n  - ".join(all_elements.keys())))



import copy
from selenium.webdriver.common.by import By

class Locator(object):

    def __init__(self, by, value):
        self.by = by
        self.value = value

    def append(self, relative_path):
        self.value += relative_path

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def get_tuple(self):
        """
        Returns a Locator as a tuple of (by, path)
        """
        return (self.by, self.value)

    def copy(self):
        return copy.deepcopy(self)

    def dict(self):
        return {
            'by': self.by,
            'value': self.value
        }

    @staticmethod
    def by_xpath(value):
        return Locator(by=By.XPATH, value=value)

    @staticmethod
    def by_css_selector(value):
        return Locator(by=By.CSS_SELECTOR, value=value)

    @staticmethod
    def by_name(value):
        return Locator(by=By.NAME, value=value)

    @staticmethod
    def by_class_name(value):
        return Locator(by=By.CLASS_NAME, value=value)

    @staticmethod
    def by_id(value):
        return Locator(by=By.ID, value=value)

    @staticmethod
    def by_link_text(value):
        return Locator(by=By.LINK_TEXT, value=value)

    @staticmethod
    def by_partial_link_text(value):
        return Locator(by=By.PARTIAL_LINK_TEXT, value=value)

    @staticmethod
    def by_tag_name(value):
        return Locator(by=By.TAG_NAME, value=value)



from seleniumpm.webelements.element import take_screenshot_on_element_error
from seleniumpm.webelements.element import Element
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException


class Clickable(Element):
    def __init__(self, driver, locator):
        super(Clickable, self).__init__(driver, locator)

    @take_screenshot_on_element_error
    def wait_for_clickable(self, timeout=None):
        timeout = timeout if timeout is not None else self.element_timeout
        try:
            WebDriverWait(self.driver, timeout).until(
                EC.element_to_be_clickable((self.locator.by, self.locator.value)))
        except TimeoutException as e:
            e.message = "TimeoutException waiting for clickable {}={} with timeout={}s ({})".format(self.locator.by,
                                                                                                    self.locator.value,
                                                                                                    timeout,
                                                                                                    self.__class__)
            e.msg = e.message
            raise e
        return self

    @take_screenshot_on_element_error
    def click(self, checkVisibility=False):
        """
        This is the standard selenium click() operation

        :param checkVisibility: [True|False] This allows you to check for visibility before the click - Default: False
        """
        if checkVisibility:
            self.is_present_and_visible()
        self.get_webelement().click()

    @take_screenshot_on_element_error
    def click_invisible(self):
        """
        This bypasses Selenium's enforcement on only allowing clicks on visible objects

        :return: self
        """
        self.driver.execute_script("arguments[0].click();", self.get_webelement())



import re
from selenium.webdriver.common.by import By
from seleniumpm.webelements.element import Element
from seleniumpm.locator import Locator


class Table(Element):
    """
    Searching and enumerating over a Table rows and columns can be complicated. This class aims at simplifying common
    tasks that can be extended to fit your application.
    """

    def __init__(self, driver, locator):
        if locator.by != By.XPATH:
            raise AttributeError("Tables only support XPATH locators")
        super(Table, self).__init__(driver, locator)

    def get_row_index(self, column_index, pattern, regex_flag=0):
        """
        Retrieves the row_index that a given 'pattern' is found. The use case is: I want to find which row that contains
        ID '12345'.
        :param column_index: The column number (starting at 0) that you want to search
        :param pattern: The pattern you want to search (e.g. "Foobar")
        :param regex_flag: These are the re flags (e.g. re.IGNORECASE) - Default: 0
        :return: The row index that contains the pattern, otherwise returns -1 if not found
        """
        elements = self.driver.find_elements(By.XPATH, "{}{}".format(
            self.locator.value, "/tbody/tr/td[{}]".format(column_index)))
        row_index = 0
        for field in elements:
            if re.search(pattern=pattern, string=field.get_text(), flags=regex_flag):
                return row_index
            row_index += 1
        return -1

    def get_column_names(self):
        """
        Retrieves a list of column names. The column name indexes will correspond to the column_index for the value on
        every row.

        NOTE: The column names assume that the table specified them under '//thead/tr/th'. If your table was not defined
        in this manner, then this operation will return an empty list
        :return:
        """
        elements = self.driver.find_elements(By.XPATH, "{}{}".format(self.locator.value, "/thead/tr/th"))
        column_names = []
        for header in elements:
            column_names.append(header.get_text())
        return column_names

    def get_element(self, row_index, column_index):
        """
        Returns a WebDriver Element reference to the specified row + column
        :param row_index: The row number (starting at 0)
        :param column_index: The column number (starting at 0)
        :return:
        :raises NoSuchElementError: if the element doesn't exist
        """
        return self.driver.find_element(By.XPATH, "{}{}".format(
            self.locator.value, "/tbody/tr[{}]/td[{}]".format(row_index, column_index)))

    def get_locator(self, row_index, column_index):
        """
        Returns a Locator object to the specified row + column
        :param row_index: The row number (starting at 0)
        :param column_index: The column number (starting at 0)
        :return:
        """
        return Locator(By.XPATH, "{}{}".format(
            self.locator.value, "/tbody/tr[{}]/td[{}]".format(row_index, column_index)))

    def get_rows(self):
        """
        Returns a list of WebDriver Elements for each row in the table
        :return:
        """
        return self.driver.find_elements(By.XPATH, "{}{}".format(self.locator.value, "/tbody/tr"))

    def count_rows(self):
        """
        Returns a count of the number of rows in the table
        :return:
        """
        return len(self.get_rows())
